/* Einartas Jonas Globis, PS 5.1 */
#include <stdio.h>
#include <stdlib.h>
#include "Btree.h"

branch CreateBranch(void* data, int label, branch top, int side, branch left, branch right)
{
    branch pointer = (branch) malloc(sizeof(struct tree));
    pointer->data = data;
    pointer->label = label;
    pointer->top = top;
    pointer->side = side;
    pointer->left = left;
    pointer->right = right;
    return pointer;
}

branch CreateTree()
{
    root pointer = (root) malloc(sizeof(struct root));
    pointer->element = CreateBranch(0, 0, NULL, 0, NULL, NULL);
    return pointer;
}

branch DeleteTree(branch pointer)
{
    if(pointer->left != NULL)
        DeleteTree(pointer->left);
    if(pointer->right != NULL)
        DeleteTree(pointer->right);
    branch top = pointer->top;
    free(pointer);
    return top;
}

branch SearchValue(int searchedvalue, branch pointer)
{
    if(searchedvalue == pointer->label)
        return pointer;
    if(searchedvalue < pointer->label)
    {
        if (pointer->left == NULL)
            return NULL;
        else return SearchValue(searchedvalue, pointer->left);
    }
    if(searchedvalue > pointer->label)
        {
        if (pointer->right == NULL)
            return NULL;
        else  return SearchValue(searchedvalue, pointer->right);
        }
    return -1;
}

int InsertValue(int insertedvalue, void *inserteddata, branch pointer)
{
    if(pointer->label == 0)
    {
        pointer->label = insertedvalue;
        pointer->data = inserteddata;
        return 1;
    }
    if(insertedvalue == pointer->label)
        return 0;
    if(insertedvalue < pointer->label)
    {
        if(pointer->left == NULL)
        {
            pointer->left = CreateBranch(inserteddata, insertedvalue, pointer, 1, NULL, NULL);
            return 1;
        }
        else
            return InsertValue(insertedvalue, inserteddata, pointer->left);
     }
    if(insertedvalue > pointer->label)
    {
        if(pointer->right == NULL)
        {
            pointer->right = CreateBranch(inserteddata, insertedvalue, pointer, 2, NULL, NULL);
            return 1;
        }
        else
            return InsertValue(insertedvalue, inserteddata, pointer->right);
    }
    return -1;
}

branch SearchHighest(branch pointer)
{
    if (pointer == NULL)
        return NULL;
    if(pointer->right == NULL)
        return pointer;
    else return SearchHighest(pointer->right);
}

branch SearchLowest(branch pointer)
{
    if (pointer == NULL)
        return NULL;
    if(pointer->left == NULL)
        return pointer;
    else return SearchLowest(pointer->left);
}

branch ExchangeBranches(branch pointer, branch pointer2)
{
    void *data = pointer->data;
    int label = pointer->label;
    pointer->data = pointer2->data;
    pointer->label = pointer2->label;
    pointer2->data = data;
    pointer2->label = label;
    return pointer2;
}

void ExtractLeaf(branch pointer)
{
    if(pointer->side == 1)
        (pointer->top)->left = NULL;
    if(pointer->side == 2)
        (pointer->top)->right = NULL;
}

void ExtractRightBranch(branch pointer)
{
    (pointer->right)->top = pointer->top;
    if(pointer->side == 1)
        (pointer->top)->left = pointer->right;
    if(pointer->side == 2)
        (pointer->top)->right = pointer->right;
}

void ExtractLeftBranch(branch pointer)
{
    (pointer->left)->top = pointer->top;
    if(pointer->side == 1)
        (pointer->top)->left = pointer->left;
    if(pointer->side == 2)
        (pointer->top)->right = pointer->left;
}

branch ExtractValue(int extractedvalue, branch pointer)
{
    if(extractedvalue == pointer->label)
    {
        if(pointer->left == NULL)
        {
            if(pointer->right == NULL)
                ExtractLeaf(pointer);
            else
                ExtractRightBranch(pointer);
        }
        else
        {
            if(pointer->right == NULL)
                ExtractLeftBranch(pointer);
            else
            {
                branch pointer2 = SearchHighest(pointer->left);
                pointer = ExchangeBranches(pointer, pointer2);
                if (pointer->left != NULL)
                    ExtractLeftBranch(pointer);
            }
        }
        pointer->top = NULL;
        pointer->side = 0;
        pointer->left = NULL;
        pointer->right = NULL;
        return pointer;
    }
    if(extractedvalue < pointer->label)
    {
        if(pointer->left == NULL)
            return NULL;
        else return ExtractValue(extractedvalue, pointer->left);
    }
    if(extractedvalue > pointer->label)
    {
        if(pointer->right == NULL)
            return NULL;
        else return ExtractValue(extractedvalue, pointer->right);
    }
    return -1;
}

int CheckFull()
{
    branch pointer = (branch) malloc(sizeof(struct tree));
    if (pointer == NULL)
    {
        free(pointer);
        return 0;
    }
    else
    {
        free(pointer);
        return 1;
    }
}

int CheckEmpty(branch pointer)
{
    return pointer == 0;
}

int FindSize(branch pointer, int size)
{
    if(pointer->left != NULL)
        size = FindSize(pointer->left, size);
    if(pointer->right != NULL)
        size = FindSize(pointer->right, size);
    return ++size;
}

branch *MakeArray(int size)
{
    return (branch*) malloc(size*sizeof(branch));
}

int FillArray(branch pointer, branch *array, int sequence)
{
    if(pointer->left != NULL)
       sequence = FillArray(pointer->left, array, sequence);
    array[sequence] = pointer;
    ++sequence;
    if(pointer->right != NULL)
        FillArray(pointer->right, array, sequence);
    return sequence;
}

void RebuildTree(branch *array, branch pointer, int low, int high, int cursize)
{
    int nextlow, nexthigh;


    if (cursize > low)
    {
        nextlow = (cursize+low)/2;
        array[nextlow-1]->top = pointer;
        pointer->left = array[nextlow-1];
        RebuildTree(array, array[nextlow-1], low, cursize, nextlow);
    }
    else pointer->left = NULL;

    if (cursize < high-1)
    {
        if((cursize+high)%2 == 1)
            nexthigh = ((cursize+high)/2)+1;
        else nexthigh = (cursize+high)/2;
        array[nexthigh-1]->top = pointer;
        pointer->right = array[nexthigh-1];
        RebuildTree(array, array[nexthigh-1], cursize+1, high, nexthigh);
    }
    else pointer->right = NULL;
}

branch BalanceTree(branch *pointer)
{
    int size = FindSize(*pointer, 0);
    branch *array = MakeArray(size);
    FillArray(*pointer, array, 0);
    int middle;
    if(size%2 == 1)
        middle = (size/2)+1;
    else middle = size/2;
    array[middle-1]->top = NULL;
    RebuildTree (array, array[middle-1], 1, size, middle);
    *pointer = array[middle-1];
    branch pointer2 = SearchHighest(*pointer);
    pointer2->right = array[size-1];
    (array[size-1])->top = pointer2;
    (array[size-1])->side = 2;
    (array[size-1])->left = NULL;
    (array[size-1])->right = NULL;
    free(array);
}

void PrintTree(branch pointer)
{
    printf("   %d\n", pointer->label);
    if(pointer->left != NULL)
    {
        printf("->left");
        PrintTree(pointer->left);
    }
    if(pointer->right != NULL)
    {
        printf("->right");
        PrintTree(pointer->right);
    }
    printf("->up");
}
